# -*- coding: utf-8 -*-
#############################
# File:    image.rb         #
# Authors:                  #
# Mattias Roback (matro414) #
# Magnus Suther (magsu191)  #
#############################


def image(func,args,object,var)
  # This function handles methods on Image objects.

  length = 0
  length = args.length if args.is_a?(Array)
  #Calls the right function depending on which is called.
  #(Functions are defined below)
  return_value = Null.new
  case func
  when "save" then save(args,object,var,length)
  when "slideshow" then slideshow(args,object,var,length)
  when "display" then display(args,object,var,length)
  when "rotate" then rotate(args,object,var,length)
  when "flip" then flip(args,object,var,length)
  when "scale" then scale(args,object,var,length)
  when "bw" then bw(args,object,var,length)
  when "addText" then add_text(args,object,var,length)
  when "watermark" then watermark(args,object,var,length)
  when "getDimensions" then return_value = get_dimensions(args,object,var,length)
  when "getFilename" then return_value = get_filename(args,object,var,length)
  when "convert" then convert(args,object,var,length)
  when "polaroid" then polaroid(args, object, var, length)
  else
		raise("<NOT_DEFINED_ERROR>: Function #{func.inspect}() not " + 
		"defined for class Image.")
  end
  return return_value
end

#################### IMAGE MANIPULATION FUNCTIONS ####################

########## Save ##########
def save(args,object,var,length)
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for "+
  "function #{var.eval()}.save().") if length < 0 or length > 1
  #If a folder is given
  begin
    # If a new folder is given as an argument
    if length == 1
      raise("<TYPE_ERROR>: Argument to #{var.eval().to_s}.save() must "+
      "be a string.") if !args[0].is_a?(String)
      FileUtils.mkdir_p(args[0])
      #If folder name does not end with "/" one is added.
      args[0] += "/" if !args[0].end_with?("/")
    end
    object.each do |image|
      dir = File.dirname(image.filename)
      if length == 1 #Writes image to directory if given.
        image.write(args[0].to_s+(image.filename.split("/")).last.to_s)
      else #Writes over the old image.
        image.write(image.filename)
      end
    end
  rescue (SystemCallError)
		raise("<IO_ERROR>: Unable to create directory #{dir.inspect}.")
  rescue (Magick::ImageMagickError)
		raise("<IO_ERROR>: Unable to save image to file #{args[0].inspect}.")
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.save() failed.")
  end
end

########## Slideshow ##########
def slideshow(args,object,var,length)
  #Creates a slideshow of the images
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for "+
  "function #{var.eval()}.slideshow().") if length != 1
  begin
    object.delay = args[0]
    object.animate
  rescue TypeError
		raise("<TYPE_ERROR>: Wrong argument type for function "+
		"#{var.eval()}.slideshow(), expected Integer.")
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.slideshow() failed.")
  end
end

########## Display ##########
def display(args,object,var,length)
  # Display all images in object, one at a time in
  # the same window. Step through images with space.
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for "+
  "function #{var.eval()}.display().") if length != 0
  begin
    object.display()
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.display() failed.")
  end
end

########## Rotate ##########
def rotate(args,object,var,length)
  # Rotate each of the images in object, and
  # put the rotated images back into object.
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for "+
  "function #{var.eval()}.rotate().") if length != 1
  begin
    0.upto(object.length-1) do |n|
      object[n].rotate!(args[0])
    end
  rescue TypeError
		raise("<TYPE_ERROR>: Wrong argument type for function "+
		"#{var.eval()}.rotate(), expected Integer or Float.")
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.rotate() failed.")
  end
end

########## Flip ##########
def flip(args,object,var,length)
  # Flip each image in object
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for "+
  "function #{var.eval()}.flip().") if length != 0
  begin
    0.upto(object.length-1) do |n|
      object[n].flip!()
    end
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.flip() failed.")
  end  
end

########## Scale ##########
def scale(args,object,var,length)
  # Scale each image in object.
  # If one argument is given, scale by percentage size change.
  # If two arguments, scale by the desired width and height.
  
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for "+
  "function #{var.eval()}.scale().") if length < 1 or length > 2
  begin
    if length == 1
			raise("<ARGUMENT_ERROR>: Argument(s) to scale() must be a "+
			"positive integer or float.") if args[0] < 0
      0.upto(object.length-1) do |n|
        object[n].scale!(args[0])
      end
    else
      raise("<ARGUMENT_ERROR>: Argument(s) to scale() must be a "+
			"positive integer or float.") if args[0] < 0 or args[1] < 0
      0.upto(object.length-1) do |n|
        object[n].scale!(args[0],args[1])
      end
    end
  rescue TypeError
		raise("<TYPE_ERROR>: Wrong argument type for "+
		"function #{var.eval()}.scale(), expected Integer or Float.")
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.scale() failed.")
  end  
end

########## Black & White ##########
def bw(args,object,var,length)
  # Convert all images in object to grayscale
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for 
  function #{var.eval()}.bw().") if length != 0
  begin
    0.upto(object.length-1) do |n|
      object[n] = object[n].quantize(256, Magick::GRAYColorspace)
    end
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.bw() failed.")
  end
end

########## Add Text ##########
def add_text(args,object,var,length)
  # Add a text on top of every image in object.
  # Default parameters is: 
  # text,x=0,y=0,pointsize=32,fill="white",font="Times",font_weight=400,
  #font_style=Magick::NormalStyle, stroke="none",stroke_width=0

  # When skipping parameters, null has to be provided as argument in 
  # place for the parameters to be skipped.

  raise("<ARGUMENT_ERROR>: Wrong number of arguments for 
  function #{var.eval()}.addText()") if length < 1 or length > 11
  
  # Set all attributes to their default values
  text = args[0]
  x = 0
  y = 0
  pointsize = 32
  fill = "white"
  font = "Times"
  font_weight = 400
  font_style = "normal"
  stroke = "transparent"
  stroke_width = 0
  
  # Set attributes to values sent as parameters, if given
  x = args[1] if !args[1].nil? and !args[1].is_a?(Null)
  y = args[2] if !args[2].nil? and !args[2].is_a?(Null)
  pointsize = args[3] if !args[3].nil? and !args[3].is_a?(Null)
  fill = args[4] if !args[4].nil? and !args[4].is_a?(Null)
  font = args[5] if !args[5].nil? and !args[5].is_a?(Null)
  font_weight = args[6] if !args[6].nil? and !args[6].is_a?(Null)
  font_style = args[7] if !args[7].nil? and !args[7].is_a?(Null)
  stroke = args[8] if !args[8].nil? and !args[8].is_a?(Null)
  stroke_width = args[9] if !args[9].nil? and !args[9].is_a?(Null)
  
  if font_style != "normal" and font_style != "italic"
		raise("<TYPE_ERROR>: Font style must be normal or italic.")
  end
  begin
    # Put text on each image in object.
    object.each do |image|
      gc = Magick::Draw.new
      gc.annotate(image, 0, 0, x, y, text) do
        self.gravity = Magick::CenterGravity
        self.pointsize = pointsize
        self.font_family = font
        self.font_style = Magick::NormalStyle
        self.font_style = Magick::ItalicStyle if font_style == "italic"
        self.fill = fill
        self.stroke = stroke
        self.stroke_width = stroke_width
        self.font_weight = font_weight
      end
    end
  rescue (TypeError)
		raise("<TYPE_ERROR>: #{var.eval()}.addText() should have parameters "+
		"(text(string),x(int),y(int),pointsize(int),fill(string),font(string),"+
		"font_weight(int),font_style(string),stroke_color(string),"+
		"stroke_width(float)) or null for default.")
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.addText() failed.")
  end
end

########## Watermark ##########
def watermark(args,object,var,length)
  # Put a watermark on each image in object.
  # A watermark always has white as font color.
  # Default parameters is:
  # x=0, y=0, pointsize=32, font='Times', font_weight=400, 
  # font_style='normal', rotation=-90
  
  raise("<ARGUMENT_ERROR>: Wrong number of arguments for "+
  "function #{var.eval()}.watermark().") if length < 1 or length > 8
  
  text = args[0]
  x = 0
  y = 0
  pointsize = 32
  font = "Times"
  font_style = "normal"
  font_weight = 400
  rotation = -90
  
  # Set attributes to values sent as parameters, if given
  x = args[1] if !args[1].nil? and !args[1].is_a?(Null)
  y = args[2] if !args[2].nil? and !args[2].is_a?(Null)
  pointsize = args[3] if !args[3].nil? and !args[3].is_a?(Null)
  font = args[4] if !args[4].nil? and !args[4].is_a?(Null)
  font_weight = args[5] if !args[5].nil? and !args[5].is_a?(Null)
  font_style = args[6] if !args[6].nil? and !args[6].is_a?(Null)
  rotation = args[7] if !args[7].nil? and !args[7].is_a?(Null)
  
  raise("<TYPE_ERROR>: Font style must be normal or 
  italic.") if font_style != "normal" and font_style != "italic"
  
  begin
    object.each_with_index do |image,index|
      # For each image, create a new image to which the text is
      # annotated, and then use this image as watermark on
      # the images in object.
      mark = Magick::Image.new(image.columns,image.rows) do
        self.background_color = 'none'
      end
      gc = Magick::Draw.new
      mark.rotate!(-rotation)
      gc.annotate(mark, 0, 0, 0, 0, text) do
        self.gravity = Magick::CenterGravity
        self.pointsize = pointsize
        self.font_family = font
        self.font_weight = font_weight
        self.font_style = Magick::NormalStyle
        self.font_style = Magick::ItalicStyle if font_style == "italic"
        self.fill = "white"
        self.stroke = "none"
      end
      mark.rotate!(rotation)
      object[index] = image.watermark(mark,0.15,0,Magick::CenterGravity,x,y)
    end
  rescue (TypeError)
		raise("<TYPE_ERROR>: #{var.eval()}.addText() should have parameters "+
		"(text(string),x(int),y(int),pointsize(int),fill(string),font(string),"+
		"font_weight(int),font_style(string),stroke_color(string),"+
		"stroke_width(float)) or null for default.")
  rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.watermark() failed.")
  end
end

########## Get Dimensions ##########
def get_dimensions(args,object,var,length)
  # Get an array with [x, y] for each image in object.
  dimensions = []
  object.each do |image|
    dimensions << [image.columns,image.rows]
  end
  return dimensions
end

########## Get Filename ##########
def get_filename(args,object,var,length)
  # Get an array with [filename] for each image in object.
  filenames = []
  object.each do |image|
    filenames << image.filename
  end
  return filenames
end

########## Convert ##########
def convert(args,object,var,length)
    # Convert each image in object to the format given as
    # argument.
    # The images will be saved with this new extension, but the converted
		# images will not be loaded back into object.
  if length != 1
		raise("<ARGUMENT_ERROR>: Wrong number of arguments "+
		"for #{var.eval()}.convert().")
  else
		raise("<TYPE_ERROR>: Wrong argument type for function "+
		"#{var.eval()}.convert(), expected String.") if !args[0].is_a?(String)
    format = args[0].downcase().gsub(".","")
    if !["jpg","jpeg","png","gif","bmp"].include?(format)
			raise("<TYPE_ERROR>: Can only convert to jpg, jpeg, png, gif or bmp.")
    end 
    object.each_with_index do |image,index|
      # Substitute the file extension and let RMagick do the
      # convert for each image. Note that this saves each image to the
      # same folder it was read from, but with a new extension.
      ext_length = File.extname(object[index].filename).length
      object[index].write(object[index].filename[0..-(ext_length)]+format)
    end
  end
end

########## Polaroid ##########
def polaroid(args,object,var,length)
	# Make all images in object polaroid.
	
	raise("<ARGUMENT_ERROR>: Wrong number of arguments "+
		"for #{var.eval()}.polaroid().") if not (length.eql?(1) or
		 length.eql?(0))
		 
	raise("<TYPE_ERROR>: Wrong argument type for function "+
		"#{var.eval()}.polaroid(), expected String") if not args.nil? and 
		args[0].is_a?(Null)
		
	# Set default value, and get the argument if given.
	text = ""
	text = args[0] if not args.nil? and not args[0].is_a?(Null)
	
	begin
		object.each_with_index do |image, index|
			object[index][:caption] = text
			object[index] = image.polaroid{self.gravity = Magick::CenterGravity}
			object[index].change_geometry!("#{image.columns}x#{image.rows}") do 
			|ncols, nrows, img|
				img.resize!(ncols, nrows)
			end
		end
	rescue
		raise("<IMAGE_ERROR>: #{var.eval()}.polaroid() failed.")
	end
end
