# -*- coding: utf-8 -*-
#############################
# File:    scopehandler.rb  #
# Authors:                  #
# Mattias Roback (matro414) #
# Magnus Suther (magsu191)  #
#############################

class Scope_handler
	# A helping hand to handle scopes. All methods are static, so we can
	# handle scopes in all files with ease, without messing with imports.
	
	# @@variables is an array containing hashes. The last hash in this
	# array is the current scope, the first is the global scope. 
	# The hash consists of keys(variable identifiers) and 
	# their evaluated value.
	@@variables = Array.new()
	@@variables << Hash.new()
	
	# @@functions is a hash, where the key is the name of the function, 
	# and the value is an array containing two elements. The first element
	# is an array containing arrays with the function parameters and their
	# associated value, and the second element contains the 
	# function block/body.
	@@functions = Hash.new()
	
	# @@classes is a hash, where the key is the name of a class.
	# The value is also a hash, with two key-value pairs. The first key is
	# "vars", and stores a hash with the instance variables. The second
	# key is "funcs", which stores a hash of the member functions. The key
	# will be the name of the function, and the value will be an array
	# with the function parameters and the class body/block.
	@@classes = Hash.new()
	
	
	########## Handle variable scopes ##########
	def Scope_handler.new_variable_scope()
		# Create a new scope, ie add a new hash to @variables
		@@variables << Hash.new()
		nil
	end
	
	def Scope_handler.insert_variable_scope(scope)
		# Append scope to @variables
		 @@variables << scope
		 nil
	end
	
	def Scope_handler.merge_variable_scopes(scope)
		# Merge scope into @@variables, scope should be an array
		 @@variables += scope
		 nil
	end
	
	def Scope_handler.pop_variable_scope(number_of_scopes=nil)
		# Pop, ie remove and return the last scope (or the number of scopes
		# given).
		if number_of_scopes.nil?() then
			return @@variables.pop()
		else
			return @@variables.pop(number_of_scopes) # Returns array
		end
	end
	
	def Scope_handler.assign_variable(id, value)
		# If the variable is assigned already, reassign it.
		define_new = true
    @@variables.reverse_each do |scope|
      if scope.has_key?(id)
        scope[id] = value
        define_new = false
        break
      end
    end
    @@variables.last()[id] = value if define_new
    nil
  end
  
  def Scope_handler.assign_variable_in_last_scope(id, value)
		# Assign the variable in the last scope only
		@@variables.last()[id] = value
		nil
  end
  
  def Scope_handler.get_variable_value(id)
		# Return the value of the variable in the closest scope possible.
		variable = nil
		@@variables.reverse_each do |scope|
			variable = scope[id] if scope.has_key?(id)
		end
		raise("<NOT_DEFINED_ERROR>: Variable #{id.inspect} not " + 
		"defined.") if variable.nil?
		return variable
  end
  
  def Scope_handler.get_number_of_variable_scopes()
		return @@variables.length()
  end
  
  def Scope_handler.print_variables()
		$stdout.puts @@variables.inspect
  end
  
  ########## Handle function scopes ##########
  def Scope_handler.function_defined(id)
		# Check if a function is defined.
		return false if @@functions[id].nil?
		return true
  end
  
  def Scope_handler.function_has_parameters(id)
		# Check if a parameter list is associated with a function
		return false if @@functions[id][0].nil?
		return true
  end
  
  def Scope_handler.get_function_parameters(id)
		# Return the function parameter list
		return @@functions[id][0]
  end
  
	def Scope_handler.get_function_block(id)
		# Return the function block
		return @@functions[id][1]
	end
	
	def Scope_handler.define_function(id, params, block)
		# Define a function
		@@functions[id] = [params, block]
		nil
	end
  
  ########## Handle class scopes ##########
  def Scope_handler.class_function_defined(class_name, id)
		# Check if a class member function is defined
		return false if @@classes[class_name]["funcs"][id].nil?
		return true
  end
  
  def Scope_handler.class_function_has_parameters(class_name, id)
		# Check if a class member function has parameters
		return false if @@classes[class_name]["funcs"][id][0].nil?
		return true
  end
  
  def Scope_handler.class_defined(class_name)
		# Check if a class is already defined
		return false if @@classes[class_name].nil?
		return true
  end
  
  def Scope_handler.get_class_function_parameters(class_name, id)
		# Return the parameters of a class function
		return @@classes[class_name]["funcs"][id][0]
  end
  
  def Scope_handler.get_class_function_block(class_name, id)
		# Return the block of a class function
		return @@classes[class_name]["funcs"][id][1]
  end
  
  def Scope_handler.get_class_variables(class_name)
		# Get the class variables defined in a class
		return @@classes[class_name]["vars"]
  end
  
  def Scope_handler.get_class(class_name)
		# Return everything associated with a class
		return @@classes[class_name]
  end
  
  def Scope_handler.define_class(class_name)
		# Define a class
		@@classes[class_name] = {"vars"=>{},"funcs"=>{}}
		nil
  end
  
  def Scope_handler.define_class_function(class_name, func_name, params, block)
		# Define a class function
		 @@classes[class_name]["funcs"][func_name] = [params, block]
		nil
  end
  
  def Scope_handler.assign_class_variable(class_name, var_name, value)
		# Assign a value to an instance variable
		@@classes[class_name]["vars"][var_name] = value
		nil
  end
end
