# -*- coding: utf-8 -*-
#############################
# File:    tree.rb          #
# Authors:                  #
# Mattias Roback (matro414) #
# Magnus Suther (magsu191)  #
#############################

# Require the right files, depending on if the language is
# installed to /usr/local/* or not.
if File.exists?("/usr/local/lib/ImgPL") then
  require '/usr/local/lib/ImgPL/scopehandler'
  require '/usr/local/lib/ImgPL/image'
else
  if not File.exists?("scopehandler.rb") then
    abort("Could not find required file 'scopehandler.rb'! Aborting.")
  else
    require 'scopehandler'
  end
	if not File.exists?("image.rb") then
		abort("Could not find required file 'image.rb'! Aborting.")
	else
		require 'image'
	end
end


class Program
  def initialize(lines)
    @lines = lines
  end
  def eval()
    return_value = nil
    return_value = @lines.eval()
    return_value = nil if return_value.is_a?(Break_stmt) or 
      return_value.is_a?(Return_stmt)
    return_value
  end
end

class Lines
  def initialize(line,lines=nil)
    @line = line
    @lines = lines
  end
  def eval()
    return_value = nil
    if @lines.nil?
      return_value = @line.eval()
    else
      return_value = @line.eval()
      return_value = @lines.eval() if !return_value.is_a?(Return_stmt) and
        !return_value.is_a?(Break_stmt)
    end
    return_value
  end
end

class Line
  def initialize(expr=nil,stmt=nil)
    @expr=expr
    @stmt=stmt
  end
  def eval()
    return_value = nil
    if !@expr.nil?
      return_value = @expr.eval()
    elsif !@stmt.nil?
      return_value = @stmt.eval()
    end
    return_value
  end
end

class Expr
  def initialize(add_expr)
    @add_expr = add_expr
  end
  def eval()
    @add_expr.eval()
  end
end

class Add_expr
  def initialize(add_expr,op=nil,multi_expr=nil)
    @add_expr=add_expr
    @op=op
    @multi_expr=multi_expr
  end
  def eval()
    if @multi_expr.nil?
      return @add_expr.eval()
    else
      add = @add_expr.eval()
      multi = @multi_expr.eval()
      case @op
        #Addition
      when "+" then
        if add.is_a?(Array)
          if multi.is_a?(Array)
            #If both is array they are combined
            add + multi
          else
            #Else, put it in the last element
            add << multi
          end
        elsif add.is_a?(String)
          add + multi.to_s
        else
          begin
            add + multi
          rescue
						raise("<TYPE_ERROR>: Can't add #{multi.inspect} to " +
						" #{add.inspect}.")
          end
        end
        #Subtraction
      when "-" then
        if add.is_a?(Array)
          if multi.is_a?(Array)
            #Removes the elements in multi from add
            add - multi
          else
            #Removes all elements thats equal to multi
            add.delete_if{|elem|elem == multi}
          end
        elsif add.is_a?(String)
          #Removes all occurences of multi in add
          add.gsub(multi.to_s,"")
        else
          begin
            add - multi
          rescue
						raise("<TYPE_ERROR>: Can't subtract #{multi.inspect} from "+ 
						"#{add.inspect}.")
          end
        end
      end
    end
  end
end

class Multi_expr
  def initialize(multi_expr,op=nil,unary_expr=nil)
    @multi_expr = multi_expr
    @op= op
    @unary_expr = unary_expr
  end
  def eval()
    if @unary_expr.nil?
      return @multi_expr.eval()
    else
      multi = @multi_expr.eval()
      unary = @unary_expr.eval()
      case @op
        #Multiplication
      when "*" then
        if unary.is_a?(Array) or unary.is_a?(String)
					raise("<TYPE_ERROR>: Can't multiply #{multi.inspect} with " + 
					"#{unary.inspect}.")
        else
          multi * unary
        end
        #Division
      when "/" then
        if (multi.is_a?(Integer) or multi.is_a?(Float)) and
            (unary.is_a?(Integer) or unary.is_a?(Float))
          begin
            multi / unary
          rescue
						raise("<ZERO_DIVISION_ERROR>: Only Chuck Norris can " + 
						"divide #{multi.inspect} by 0.")
          end
        else
					raise("<TYPE_ERROR>: Can't divide #{multi.inspect} by " + 
					"#{unary.inspect}.")
        end
        #Modulo
      when "%" then
        if (multi.is_a?(Integer) or multi.is_a?(Float)) and
            (unary.is_a?(Integer) or unary.is_a?(Float))
          multi % unary
        else
					raise("<TYPE_ERROR>: Can't use modulo for #{multi.inspect} " + 
					"and #{unary.inspect}.")
        end
      end
    end
  end
end

class Unary_expr
  def initialize(unary_expr,op=nil)
    @op = op
    @unary_expr = unary_expr
  end
  def eval()
    if @op.nil?
      return @unary_expr.eval()
    else
      unary = @unary_expr.eval()
      if unary.is_a?(Integer) or unary.is_a?(Float)
        case @op
        when "+" then +unary
        when "-" then -unary
        end
      else
				raise("<TYPE_ERROR>: Can't use #{@op.inspect} for " + 
				"#{unary.inspect}.")
      end
    end
  end
end

class Power_expr
  def initialize(expr,power_expr=nil)
    @power_expr = power_expr
    @expr = expr
  end
  def eval()
    if @power_expr.nil?
      return @expr.eval()
    else
      power = @power_expr.eval()
      expr = @expr.eval()
      if power.is_a?(Integer) or power.is_a?(Float) and
        expr.is_a?(Integer) or expr.is_a?(Float)
        return power ** expr
      else
				raise("<TYPE_ERROR>: Can't calculate #{power.inspect} ** " + 
				"#{expr.inspect}.")
      end
    end
  end
end

class Rel_expr
  def initialize(expr1,op = nil,expr2 = nil)
    @expr1 = expr1
    @op = op
    @expr2 = expr2
  end
  def eval()
    expr1 = @expr1.eval()
    if !@expr2.nil?
      expr2 = @expr2.eval()
      len1 = expr1.length if !expr1.eql?(true) and !expr1.eql?(false)
      len2 = expr2.length if !expr2.eql?(true) and !expr2.eql?(false)

			raise("<TYPE_ERROR>: Operator <,>,<=,>= not " + 
			"defined for boolean.") if ["<",">","<=",">="].include?(@op) and 
			(len1.nil? or len2.nil?)
			
      # Compares length or objects depending on the operator
      case @op
      when "<" then len1 < len2
      when ">" then len1 > len2
      when "<=" then len1 <= len2
      when ">=" then len1 >= len2
      when "is" then expr1 == expr2
      when "is not" then expr1 != expr2
      when "in" then expr2.include? expr1
      when "not in" then !expr2.include? expr1
      end
    else
      return expr1
    end
  end
end

class Datatypes
  def initialize(value)
    @value = value
  end
  def eval()
    return @value.eval()
  end
end

class Int_type
  def initialize(value)
    @value = value
  end
  def eval()
    @value
  end
end

class Float_type
  def initialize(value)
    @value = value
  end
  def eval()
    @value
  end
end

class String_type
  def initialize(value)
    @value = value
    @value = value[1..-2] if value =~ /\"/
  end
  def eval()
    @value
  end
end

class Bool_type
  def initialize(value)
    if value == "true"
      @value = true
    else
      @value = false
    end
  end
  def eval()
    @value
  end
end

class Range_type
  def initialize(s,e)
    @start = s
    @end = e
  end
  def eval()
    return (@start..@end).to_a
  end
end

class Array_type
  def initialize(elements = nil)
    @elements = elements
  end
  def eval()
    @elements = @elements.eval() if !@elements.nil?
    @elements = [] if @elements.nil?
    @elements
  end
end

class Elements
  def initialize(expr,elements=nil)
    @expr = expr
    @elements = elements
  end
  def eval()
    if !@elements.nil?
      a = @elements.eval()
    else
      a = []
    end
    a << @expr.eval()
  end
end

class Identifier
  def initialize(var_name)
    @var_name = var_name
  end
  def eval()
    @var_name
  end
end

class Null
  def eval()
    Null.new()
  end
end

class Class_var
  def initialize(var_name)
    @var_name = var_name
  end
  def eval()
    @var_name
  end
end

class Stmt
  def initialize(stmt)
    @stmt = stmt
  end
  def eval()
    return_value = nil
    return_value = @stmt.eval()
    return_value
  end
end

class Assign_stmt
  attr_accessor :identifier, :expr
  def initialize(id,expr)
    @identifier = id
    @expr = expr
  end
  def eval()
    id = @identifier.eval()    
    Scope_handler.assign_variable(id, @expr.eval())
  end
end

class Print_stmt
  def initialize(type,expr)
    @type = type
    @expr = expr
  end
  def eval()
    expr = @expr.eval()
    if expr.is_a?(Hash)
      #Only happens if expression is a class defined by user.
      str = "Class:\t#{expr["class"]}"
      str += "\nVars: "
      expr["vars"].each{|key,value|str+="\t#{key} = #{value.inspect}\n"}
      expr = str
    elsif expr.is_a?(Magick::ImageList)
      #Happens if an Image is printed
      str = "Class:\tImage\n"
      expr.each{|image|str+="#{image.filename()}\n"}
      expr = str
    end
    case @type
    when "println" then $stdout.puts expr
    when "print" then $stdout.print expr
    when "errprint" then $stderr.puts "<ERRPRINT> #{expr}"
    end
    nil
  end
end

class Get_stmt
  def initialize(id)
    @id = id
  end
  def eval()
    in_value = $stdin.gets.chomp
    value = String_type.new(in_value)
    case in_value
    when /\A\d+\.\d+\z/ then value = Float_type.new(in_value.to_f)
    when /\A\d+\z/ then value = Int_type.new(in_value.to_i)
    end
    Assign_stmt.new(@id,value).eval()
  end
end

class Get_var
  def initialize(identifier)
    @identifier = identifier
  end
  def eval()
    variable = nil
    id = @identifier.eval()
    variable = Scope_handler.get_variable_value(id)
    variable
  end
end

class If_else_stmt
  def initialize(if_stmt,else_if_stmts=nil,else_stmt=nil)
    @if_stmt = if_stmt
    @else_if_stmts = else_if_stmts
    @else_stmt = else_stmt
  end
  def eval()
    return_value = @if_stmt.eval()
    continue_else = true
    if return_value == true and !@else_if_stmts.nil?
      continue_else = @else_if_stmts.eval()
    end
    if continue_else == true and return_value == true and !@else_stmt.nil?
      return_value = @else_stmt.eval()
    end
    return_value = nil if return_value == true
    return_value = continue_else if continue_else.is_a?(Break_stmt) or
      continue_else.is_a?(Return_stmt)
    return_value
  end
end

class If_stmt
  def initialize(cond,block)
    @cond = cond
    @block = block
  end
  def eval()
    return_value = true
    Scope_handler.new_variable_scope()
    
    if(@cond.eval())
      return_value = @block.eval()
    end
    
    Scope_handler.pop_variable_scope()
    return_value
  end
end

class Else_if_stmts
  def initialize(else_if_stmt,else_if_stmts=nil)
    @else_if_stmt = else_if_stmt
    @else_if_stmts = else_if_stmts
  end
  def eval()
    return_value = true
    return_value = @else_if_stmt.eval()
    if(!@else_if_stmts.nil? and return_value==true)
      return_value = @else_if_stmts.eval()
    end
    return_value
  end
end

class Else_if_stmt
  def initialize(cond,block)
    @cond = cond
    @block = block
  end
  def eval()
    return_value = true
    Scope_handler.new_variable_scope()
    
    if(@cond.eval())
      return_value = @block.eval()
    end
    
    Scope_handler.pop_variable_scope()
    return_value
  end
end

class Else_stmt
  def initialize(block)
    @block = block
  end
  def eval()
    Scope_handler.new_variable_scope()
    return_value = @block.eval()
    Scope_handler.pop_variable_scope()
    return_value
  end
end

class While_stmt
  def initialize(cond,block)
    @cond = cond
    @block = block
  end
  def eval()
    return_value = nil
    Scope_handler.new_variable_scope()
    
    while(@cond.eval())
      return_value = @block.eval()
      break if return_value.is_a?(Break_stmt) or 
      return_value.is_a?(Return_stmt)
    end
    
    Scope_handler.pop_variable_scope()
    return_value = nil if return_value.is_a?(Break_stmt)
    return_value
  end
end

class For_stmt
  def initialize(id,op,data,block)
    @id = id.eval()
    @op = op
    @data = data.eval()
    @block = block
    case @data
    when Bignum,Fixnum then @data = @data.split()
    when Float then @data = @data.split()
    when String then 
			# If the string contains spaces, we loop word by word, 
			# otherwise we loop letter by letter.
      if @data =~ /\s/
        @data = @data.split(/\s/)
      else
        @data = @data.split(//)
      end
    end
  end
  def eval()
    return_value = nil
    Scope_handler.new_variable_scope()
    
    for i in @data
      Scope_handler.assign_variable_in_last_scope(@id, i)
      return_value = @block.eval()
      break if return_value.is_a?(Break_stmt) or return_value.is_a?(Return_stmt)
    end
    
    Scope_handler.pop_variable_scope()
    return_value = nil if return_value.is_a?(Break_stmt)
    return_value
  end
end

class Block
  def initialize(lines)
    @lines = lines
  end
  def eval()
    return_value = nil
    return_value = @lines.eval()
    return_value
  end
end

class Condition
  def initialize(cond)
    @cond = cond
  end
  def eval()
    return_value = @cond.eval()
    
    if return_value != false and return_value != true
			raise("<TYPE_ERROR>: Condition #{return_value.inspect} cannot " + 
			"be resolved to boolean.")
    end
    
    return_value
  end
end

class Or_expr
  def initialize(expr1,expr2 = nil)
    @expr1 = expr1
    @expr2 = expr2
  end
  def eval()
    if @expr2.nil?
      return @expr1.eval()
    else
      return @expr1.eval() || @expr2.eval()
    end
  end
end

class And_expr
  def initialize(expr1,expr2 = nil)
    @expr1 = expr1
    @expr2 = expr2
  end
  def eval()
    if @expr2.nil?
      return @expr1.eval()
    else
      return @expr1.eval() && @expr2.eval()
    end
  end
end

class Not_expr
  def initialize(expr,flag = nil)
    @expr = expr
    @flag = flag
  end
  def eval()
    if @flag.nil?
      return @expr.eval()
    else
      return !@expr.eval()
    end
  end
end

class Param_list
  def initialize(params, param)
    @params = params
    @param = param
  end
  def eval()
    if not @params.nil?
      a = @params.eval()
    else
      a = []
    end
    a << @param.eval()
  end
end

class Param
  def initialize(stmt)
    @stmt = stmt
  end
  def eval()
    if @stmt.is_a?(Assign_stmt)
      id = @stmt.identifier.eval()
      expr = @stmt.expr.eval()
      [id,expr]
    else
      [@stmt.eval(),nil]
    end
  end
end

class Func_call
  def initialize(id,args=nil,class_name=nil)
    @id = id
    @args = args
    @class_name = class_name
  end
  def eval()
    id = @id.eval()
    params = []
    if @class_name.nil?
    
      raise("<NOT_DEFINED_ERROR>: Function \"#{id}()\" not " + 
      "defined.") if not Scope_handler.function_defined(id)
      
      # Marshal is used to do a "deep copy" of an object.
      if Scope_handler.function_has_parameters(id) then
								func_params = Scope_handler.get_function_parameters(id)
        params = Marshal.load(Marshal.dump(func_params))
      end
      
    else
			function_defined = Scope_handler.class_function_defined(@class_name, id)
			raise("<NOT_DEFINED_ERROR>: Function \"#{id}()\" "+
			"not defined for class #{@class_name.inspect}.") if not function_defined

			if Scope_handler.class_function_has_parameters(@class_name, id) then
				func_params = Scope_handler.get_class_function_parameters(@class_name, id)
				params = Marshal.load(Marshal.dump(func_params))
			end
			
    end
    if !@args.nil?
      args = @args.eval()
      raise("<ARGUMENT_ERROR>: Wrong number of "+
      "arguments for function \"#{id}()\".") if args.length > params.length
      args.each_with_index{|elem,i|params[i][1] = elem if !elem.is_a?(Null)}
      # Checks if a parameter is missing from the user
      params.each{|e| raise("<ARGUMENT_ERROR>: Wrong number "+
      "of arguments for function \"#{id}()\".") if e[1].nil?}
    end
    # Maps function parameters together with their identifier and
    # puts them last in the variables scope.
    scope = {}
    params.map{|key,value|scope[key] = value}
				scope.each{|a,e| raise("<ARGUMENT_ERROR>: Wrong number "+
      "of arguments for function \"#{id}()\".") if e.nil?}

    # Makes sure that a function cannot use 
    # variables from other scopes (only global).
    number_of_scopes = Scope_handler.get_number_of_variable_scopes()
    tmp_variables = Scope_handler.pop_variable_scope(number_of_scopes - 1)
    Scope_handler.insert_variable_scope(tmp_variables.pop) if !@class_name.nil?
    Scope_handler.insert_variable_scope(scope)

    if @class_name.nil?
      #If the function does not belong to a class
      return_value = Scope_handler.get_function_block(id).eval()
    else
      #If the function is a class function.
      return_value = Scope_handler.get_class_function_block(@class_name, id).eval()
    end
    return_value = return_value.value.eval() if return_value.is_a?(Return_stmt)
    Scope_handler.pop_variable_scope()
    
    # Puts the other scopes variables back into the variables scope
    tmp_variables << Scope_handler.pop_variable_scope() if !@class_name.nil?
    Scope_handler.merge_variable_scopes(tmp_variables)
    
    return_value = Null.new if return_value.nil?
    return_value
  end
end

class Arg_list
  def initialize(arg,args=nil)
    @arg = arg
    @args = args
  end
  def eval()
    if not @args.nil?
      array = @args.eval()
    else
      array = []
    end
    array << @arg.eval()
  end
end

class Arg
  def initialize(arg)
    @arg = arg
  end
  def eval()
    @arg.eval()
  end
end

class Break_stmt
  def eval()
    self
  end
end

class Return_stmt
  attr_accessor :value
  def initialize(value)
    @value = value
  end
  def eval()
    self
  end
end

class Subscript_expr
  def initialize(expr,index)
    @expr = expr
    @index = index
  end
  def eval()
    expr = @expr.eval()
    index = @index.eval()
    if !index.is_a?(Integer)
			raise("<TYPE_ERROR>: Index must be an integer, found " + 
			"#{index.inspect}.")
    elsif !expr.is_a?(String) and !expr.is_a?(Array)
			raise("<TYPE_ERROR>: Can't use subscript operator for " +
			"#{expr.inspect}.")
    else
      if (index < expr.length) and (index >= -expr.length)
        if expr.is_a?(Array)
          expr[index]
        else
          expr[index..index]
        end
      else
				raise("<INDEX_OUT_OF_RANGE_ERROR>: Index #{index.inspect} " + 
				"for #{expr.inspect} is out range.")
      end
    end
  end
end

class Class_def
  def initialize(id,funcs,vars=nil)
    @id = id
    @funcs = funcs
    @vars = vars
  end
  def eval()
    id = @id.eval()
		raise("<DOUBLE_DEFINITION_ERROR>: Class #{id.inspect} already " + 
		"defined.") if Scope_handler.class_defined(id) or id.eql?("Image")
		
    Scope_handler.define_class(id)
    
    @funcs.eval(id)
    @vars.eval(id) if !@vars.nil?
  end
end

class Class_vars_def
  def initialize(var,vars=nil)
    @var = var
    @vars = vars
  end
  def eval(id)
    @var.eval(id)
    @vars.eval(id) if !@vars.nil?
    nil
  end
end

class Class_name
  def initialize(name)
    @name = name
  end
  def eval()
    @name
  end
end

class Class_var_def
  def initialize(var)
    @var = var
  end
  def eval(class_name)
    var = @var.eval()
    Scope_handler.assign_class_variable(class_name, var, nil)
  end
end

class Class_funcs_def
  def initialize(func,funcs=nil)
    @func = func
    @funcs = funcs
  end
  def eval(id)
    @func.eval(id)
    @funcs.eval(id) if !@funcs.nil?
    nil
  end
end

class Func_def
  def initialize(id, params, block)
    @id = id
    @params = params
    @block = block
  end
  def eval(class_name=nil)
    id = @id.eval()
    params = nil
    params = @params.eval() if !@params.nil?
    if class_name.nil?
      Scope_handler.define_function(id, params, @block)
    elsif
      Scope_handler.define_class_function(class_name, id, params, @block)
    end
    nil
  end
end

class Class_call
  def initialize(class_name,func_name,args=nil)
    @class_name = class_name
    @func_name = func_name
    @args = args
  end
  def eval()
    name = @func_name
    func_name = @func_name.eval()
    class_name = @class_name.eval()

    #If a function is called on an image-object.
    if class_name.eql?("Image")
      raise("<MEMBER_FUNCTION_CALL_ERROR>: Must create new instance "+
      "before calling member " + 
      "function \"#{func_name}()\".") if not func_name.eql?("new")
      args = @args.eval()
      imagelist = Magick::ImageList.new()
      
      # Searches for images in the given directories
      #-Block begin-
      args.each do |arg|
        arg = arg[0..-2] if arg[-1..-1].eql?("/")
        filetypes = [".jpg",".jpeg",".gif",".bmp",".png"]
        if File.exists?(arg) and File.directory?(arg) then
          # Add all images in arg to imagelist, non-recursive
          entries = Dir.entries(arg)
          entries.each do |ent|
            file = arg + "/" + ent
            # Opens images with the correct filetype.
            if filetypes.include?(File.extname(ent).downcase)
              imagelist.read(file) 
            end
          end
        else
          raise("<IO_ERROR>: File #{arg.inspect} not found.") if 
          !File.exists?(arg)
          
          if !(filetypes.include?(File.extname(arg).downcase))
						raise("<IO_ERROR>: #{arg.inspect} is not a valid filetype.")
            return Null.new
          else
            imagelist.read(arg)
          end
          
        end
      end
      #-Block end-
      return imagelist

      #If function is called on another object.
    else
      if func_name.eql?("new")
        scope = Scope_handler.get_class_variables(class_name)
        Scope_handler.insert_variable_scope(scope.clone())
        
        Func_call.new(name,@args,class_name).eval()
      else
				raise("<MEMBER_FUNCTION_CALL_ERROR>: Must create new instance "+
				"before calling member function \"#{func_name}()\".")
      end
      return {"class"=>class_name, "vars"=>Scope_handler.pop_variable_scope()}
    end
  end
end

class Instance_func_call
  def initialize(var_name,func_name,args=nil)
    @var_name = var_name
    @func_name = func_name
    @args = args
  end
  def eval()
    var = Get_var.new(@var_name).eval()
    if var.is_a?(Magick::ImageList)
      func = @func_name.eval()
      args = @args.eval() if !@args.nil?
      return image(func,args,var,@var_name)
    else
      class_name = var["class"] if var.is_a?(Hash)
      class_hash = Scope_handler.get_class(class_name)
      if !class_name.nil?
        Scope_handler.insert_variable_scope(var["vars"])
        return_value = Func_call.new(@func_name,@args,class_name).eval()
        Scope_handler.pop_variable_scope()
      else
				raise("<NO_INSTANCE_ERROR>: Variable " +
				"#{@var_name.eval().inspect} is not a class instance.")
      end
    end
    return_value
  end
end

class Image
  def initialize(image)
    @image = image
  end
  def eval()
    @image
  end
end
